#ifndef RK_DNSPACK_BASISCOMPOSITION_H
#define RK_DNSPACK_BASISCOMPOSITION_H

#include <set>
#include <map>
#include <stdexcept>

namespace rk
{
	namespace dnspack
	{
		class BasisComposition
		{
		public:
			BasisComposition();
			BasisComposition(const std::set<double>& radii, const std::map<double, int>& numbers);
			BasisComposition(std::set<double>&& radii, std::map<double, int>&& numbers);
			virtual ~BasisComposition();

		public:
			BasisComposition(const BasisComposition&) = default;
			BasisComposition(BasisComposition&&) noexcept = default;
			BasisComposition& operator=(const BasisComposition&) = default;
			BasisComposition& operator=(BasisComposition&&) noexcept = default;

			bool operator==(const BasisComposition& other) const;
			bool operator!=(const BasisComposition& other) const;

		public:
			std::size_t getSphereKindNumber() const noexcept;
			std::size_t getSpheresNumber() const;

			const std::set<double> getRadii() const noexcept;
			const std::map<double, int> getNumbers() const noexcept;

			bool setComposition(const std::set<double>& radii, const std::map<double, int>& numbers);
			bool setComposition(std::set<double>&& radii, std::map<double, int>&& numbers) noexcept;
			void clear() noexcept;

		private:
			std::set<double> m_radii;
			std::map<double, int> m_numbers;
		};
	}
}


inline bool rk::dnspack::BasisComposition::operator==(const BasisComposition& other) const
{
	return (m_numbers == other.m_numbers && m_radii == other.m_radii);
}

inline bool rk::dnspack::BasisComposition::operator!=(const BasisComposition& other) const
{
	return !(*this == other);
}

inline std::size_t rk::dnspack::BasisComposition::getSphereKindNumber() const noexcept
{
	return m_radii.size();
}

inline std::size_t rk::dnspack::BasisComposition::getSpheresNumber() const
{
	std::size_t spheresNumber = 0;

	for (const auto& num : m_numbers)
	{
		spheresNumber += num.second;
	}

	return spheresNumber;
}

inline const std::set<double> rk::dnspack::BasisComposition::getRadii() const noexcept
{
	return m_radii;
}

inline const std::map<double, int> rk::dnspack::BasisComposition::getNumbers() const noexcept
{
	return m_numbers;
}

inline bool rk::dnspack::BasisComposition::setComposition(const std::set<double>& radii, const std::map<double, int>& numbers)
{
	if (radii.size() != numbers.size())
	{
		return false;
	}

	m_numbers = numbers;
	m_radii = radii;

	return true;
}

inline bool rk::dnspack::BasisComposition::setComposition(std::set<double>&& radii, std::map<double, int>&& numbers) noexcept
{
	if (radii.size() != numbers.size())
	{
		return false;
	}

	m_numbers = std::move(numbers);
	m_radii = std::move(radii);

	return true;
}

inline void rk::dnspack::BasisComposition::clear() noexcept
{
	m_radii.clear();
	m_numbers.clear();
}


#endif // !RK_DNSPACK_BASISCOMPOSITION_H
