#include "ExploreManager.h"

#include <vector>
#include <thread>
#include <mutex>

#include "MakeDirectory.h"
#include "ReportCurrentSituation.h"
#include "MakeResultFile.h"
#include "CalculationAnalysis.h"

using namespace rk::dnspack;

ExploreManager::ExploreManager(const BasisComposition& basisComposition, const ExploreManager::ExploreParameters& exploreParameters)
	: m_basisComposition(basisComposition)
	, m_exploreParameters(exploreParameters)
{
}

ExploreManager::~ExploreManager()
{
}

int ExploreManager::execute()
{
	PackingFactory::OptimizeParameters optimizeParameters;
	PackingFactory::DirectoryNames directoryNames;


// ******************************************************************************************************************************************
	// initialize optimizeParameters

	optimizeParameters.initialExpandMaxChangeSize = m_exploreParameters.initialExpandMaxChangeSize;
	optimizeParameters.initialExpandMaxIteratorNumber = m_exploreParameters.initialExpandMaxIteratorNumber;
	optimizeParameters.initialArrangeMaxChangeSizeUpper = m_exploreParameters.initialArrangeMaxChangeSizeUpper;
	optimizeParameters.initialArrangeMaxChangeSizeLower = m_exploreParameters.initialArrangeMaxChangeSizeLower;
	optimizeParameters.initialArrangeIteratorNumber = m_exploreParameters.initialArrangeIteratorNumber;
	optimizeParameters.spherePackingOptimizeMaxChangeSizeUpper = m_exploreParameters.spherePackingOptimizeMaxChangeSizeUpper;
	optimizeParameters.spherePackingOptimizeMaxChangeSizeLower = m_exploreParameters.spherePackingOptimizeMaxChangeSizeLower;
	optimizeParameters.spherePackingOptimizeDecreaseFactor = m_exploreParameters.spherePackingOptimizeDecreaseFactor;
	optimizeParameters.spherePackingOptimizeIteratorNumber = m_exploreParameters.spherePackingOptimizeIteratorNumber;

	// initialize optimizeParameters
// ******************************************************************************************************************************************


// ******************************************************************************************************************************************
	// initialize directoryNames

	MakeDirectory makeDatFilesDirectory(m_exploreParameters.outputDirectoryName, "DatFiles/");
	makeDatFilesDirectory();

	MakeDirectory makeXsfFilesDirectory(m_exploreParameters.outputDirectoryName, "XsfFiles/");
	makeXsfFilesDirectory();

	directoryNames.datFilesDirectoryName = m_exploreParameters.outputDirectoryName + "DatFiles/";
	directoryNames.xsfFilesDirectoryName = m_exploreParameters.outputDirectoryName + "XsfFiles/";

	// initialize directoryNames
// ******************************************************************************************************************************************


// ******************************************************************************************************************************************
	// initialize report files

	ReportCurrentSituation reporter;
	reporter.setResultFileName(m_exploreParameters.outputDirectoryName + "currentSituation.txt");
	{
		std::stringstream reportMessage;
		reportMessage << "Thread Number  =  " << m_exploreParameters.threadNumber << "\n\n";
		reporter.report(reportMessage.str());
	}

	MakeResultFile makeResultFile;
	makeResultFile.setResultFileName(m_exploreParameters.outputDirectoryName + "result.txt");

	// initialize report files
// ******************************************************************************************************************************************


	PackingFactory::FactoryAttribute factoryAttribute;
	{
		factoryAttribute.maxTrialNumber = m_exploreParameters.maxTrialNumber;
		factoryAttribute.thresholdTrialNumber = m_exploreParameters.thresholdTrialNumber;
		factoryAttribute.createdNumber = 0;
		factoryAttribute.creatPlanNumber = m_exploreParameters.thresholdTrialNumber;

		factoryAttribute.maxFillingFactor = 0.0;
		factoryAttribute.outputMinimumFillingFactor = m_exploreParameters.outputMinimumFillingFactor;
	}


// ******************************************************************************************************************************************
	// packing produce

	std::vector<std::thread> produceTasks(m_exploreParameters.threadNumber);

	std::vector<PackingFactory> packingFactories;
	packingFactories.reserve(static_cast<std::size_t>(m_exploreParameters.threadNumber));

	for (int index = 0; index < m_exploreParameters.threadNumber; ++index)
	{
		PackingFactory packingFactory(m_basisComposition, optimizeParameters, directoryNames, factoryAttribute);
		packingFactories.push_back(packingFactory);
	}

	PackingFactory::MutexCluster mutexCluster;

	for (int index = 0; index < m_exploreParameters.threadNumber; ++index)
	{
		std::thread produceTask{ packingFactories[index], std::ref(mutexCluster) };
		produceTasks[index] = std::move(produceTask);
	}

	for (auto& produceTask : produceTasks)
	{
		produceTask.join();
	}

	// packing produce
// ******************************************************************************************************************************************


// ******************************************************************************************************************************************
	// make result file

	if (6 < m_basisComposition.getSphereKindNumber())
	{
		throw std::runtime_error("The kind of spheres is more than 6");
	}

	std::vector<std::string> sphereNames(6);
	{
		sphereNames[0] = "H";
		sphereNames[1] = "Na";
		sphereNames[2] = "Sc";
		sphereNames[3] = "Y";
		sphereNames[4] = "La";
		sphereNames[5] = "Hf";
	}

	std::vector<double> radii;
	{
		for (const auto& radius : m_basisComposition.getRadii())
		{
			radii.push_back(radius);
		}
	}

	if (6 < radii.size())
	{
		throw std::runtime_error("The kind of sphere is larger than 6.");
	}

	std::stringstream message;
	message << "#\n" << "# Spheres Information\n" << "#\n\n\n";
	message << "<Definition.of.Sphere.Species\n";

	for (std::size_t label = 0; label < m_basisComposition.getSphereKindNumber(); ++label)
	{
		message << " " << sphereNames[label] << "    " << radii[label] << "\n";
	}
	message << "Definition.of.Sphere.Species>\n\n";

	message << "Atoms.Number   " << m_basisComposition.getSpheresNumber() << "\n";
	message << "<Spheres.Composition.Number.Per.UnitCell\n";
	{
		std::size_t label = 0;

		for (const auto& number : m_basisComposition.getNumbers())
		{
			message << " " << sphereNames[label] << "    " << number.second << "\n";
			++label;
		}
	}
	message << "Spheres.Composition.Number.Per.UnitCell>\n";

	message << "\n\n#\n" << "# Optimize Paramaters\n" << "#\n\n\n";
	message << "InitialExpand.MaxChangeSize                   " << optimizeParameters.initialExpandMaxChangeSize << "\n";
	message << "InitialExpand.Maximum.Iterator.Number         " << optimizeParameters.initialExpandMaxIteratorNumber << "\n";
	message << "InitialArrange.MaxChangeSize.Upper            " << optimizeParameters.initialArrangeMaxChangeSizeUpper << "\n";
	message << "InitialArrange.MaxChangeSize.Lower            " << optimizeParameters.initialArrangeMaxChangeSizeLower << "\n";
	message << "InitialArrange.Iterator.Number                " << optimizeParameters.initialArrangeIteratorNumber << "\n";
	message << "SpherePackingOptimize.MaxChangeSize.Upper     " << optimizeParameters.spherePackingOptimizeMaxChangeSizeUpper << "\n";
	message << "SpherePackingOptimize.MaxChangeSize.Lower     " << optimizeParameters.spherePackingOptimizeMaxChangeSizeLower << "\n";
	message << "SpherePackingOptimize.Decrease.Factor         " << optimizeParameters.spherePackingOptimizeDecreaseFactor << "\n";
	message << "SpherePackingOptimize.Iterator.Number         " << optimizeParameters.spherePackingOptimizeIteratorNumber << "\n";

	message << "\n\n#\n" << "# Result Infomation\n" << "#\n\n\n";


	message << "Created.Number          = " << factoryAttribute.createdNumber << "\n";
	message << "Maximum.FillingFactor   = " << factoryAttribute.maxFillingFactor << "\n\n";

	CalculationAnalysis calculationAnalysis;

	message << "InitialArrange.MinusVolume.Reject.Number = " << calculationAnalysis.getRejectNumbers().initialArrangeMinusVolumeRejectNumber << "\n";
	message << "SpherePackingOptimize.MinusVolume.Reject.Number = " << calculationAnalysis.getRejectNumbers().spherePackingOptimizeMinusVolumeRejectNumber << "\n";
	message << "SpherePackingOptimize.Overlap.Reject.Number = " << calculationAnalysis.getRejectNumbers().spherePackingOptimizeOverlapRejectNumber << "\n";
	message << "SpherePackingOptimize.Zvalue.Reject.Number = " << calculationAnalysis.getRejectNumbers().spherePackingOptimizeZvalueRejectNumber << "\n\n\n";

	message << "#\n" << "# Filling Factor Distribution : filling factor range and the number of structures\n" << "#\n\n\n";

	for (int label = 0; label < 25; ++label)
	{
		double minFillingFactor = 0.60 + 0.01 * static_cast<double>(label);
		double supFillingFactor = minFillingFactor + 0.01;

		message << minFillingFactor << " - " << supFillingFactor << " : " << calculationAnalysis.fillingFactorDistribution[label] << "\n";
	}

	makeResultFile.reportMessage(message.str());

	// make result file
// ******************************************************************************************************************************************


	calculationAnalysis.clear();

	return 0;
}
