#include "MakeDatFile.h"

#include <map>

using namespace rk::dnspack;


MakeDatFile::MakeDatFile(const std::string& fileName, PackingStructure& packingStructure)
	: m_fileName{ fileName }
	, m_packingStructure(packingStructure)
	, m_basisComposition(packingStructure.getBasisComposition())
{
}

MakeDatFile::~MakeDatFile()
{
}

void MakeDatFile::writePackingInfo()
{
	std::size_t sphereKindNumber = m_packingStructure.getSphereKindNumber();

	std::vector<std::string> sphereNames(6);
	{
		sphereNames[0] = "H";
		sphereNames[1] = "Na";
		sphereNames[2] = "Sc";
		sphereNames[3] = "Y";
		sphereNames[4] = "La";
		sphereNames[5] = "Hf";
	}

	std::vector<double> radii;
	{
		for (const auto& radius : m_basisComposition.getRadii())
		{
			radii.push_back(radius);
		}
	}

	if (6 < radii.size())
	{
		throw std::runtime_error("The kind of sphere is larger than 6.");
	}

	std::map<double, std::string> radiiAndName;
	{
		std::vector<std::string>::const_iterator nameIter = sphereNames.begin();

		for (const auto& radius : radii)
		{
			radiiAndName.emplace(radius, *nameIter);
			++nameIter;
		}
	}

	m_textStream << "#\n" << "# Spheres Infomation\n" << "#\n\n\n";
	m_textStream << "<Definition.of.Sphere.Species\n";

	for (std::size_t label = 0; label < sphereKindNumber; ++label)
	{
		m_textStream << " " << sphereNames[label] << "    " << radii[label] << "\n";
	}
	m_textStream << "Definition.of.Sphere.Species>\n\n";

	m_textStream << "Atoms.Number   " << m_packingStructure.getSpheresNumber() << "\n";
	m_textStream << "<Spheres.Composition.Number.Per.UnitCell\n";
	{
		std::size_t label = 0;

		for (const auto& number : m_basisComposition.getNumbers())
		{
			m_textStream << " " << sphereNames[label] << "    " << number.second << "\n";
			++label;
		}
	}
	m_textStream << "Spheres.Composition.Number.Per.UnitCell>\n\n\n";


	m_textStream << "#\n" << "# Packing Structure Infomation\n" << "#\n\n\n";

	m_textStream << "Atoms.SpeciesAndCoordinates.Unit  FRAC\n<Atoms.SpeciesAndCoordinates" << "\n";
	{
		int num = 1;

		for (auto& sphere : m_packingStructure.getBasis())
		{
			m_textStream << "  " << num << "  " << radiiAndName.at(sphere.getRadius()) << "  ";
			m_textStream << sphere.getFractionalCoordinate().a << "  " << sphere.getFractionalCoordinate().b << "  " << sphere.getFractionalCoordinate().c << "\n";

			++num;
		}
	}
	m_textStream << "Atoms.SpeciesAndCoordinates>" << "\n\n\n";

	const BravaisLattice& bravaisLattice = m_packingStructure.getBravaisLattice();

	m_textStream << "Atoms.UnitVectors.Unit   Ang\n<Atoms.UnitVectors\n";
	m_textStream << "  " << bravaisLattice[0].x << "  " << bravaisLattice[0].y << "  " << bravaisLattice[0].z << "\n";
	m_textStream << "  " << bravaisLattice[1].x << "  " << bravaisLattice[1].y << "  " << bravaisLattice[1].z << "\n";
	m_textStream << "  " << bravaisLattice[2].x << "  " << bravaisLattice[2].y << "  " << bravaisLattice[2].z << "\n";
	m_textStream << "Atoms.UnitVectors>" << "\n\n\n";

	double overlapVolume = 0.0;
	double minimumZ = 0.0;
	{
		if (m_packingStructure.overlapExistence(overlapVolume, minimumZ))
		{
			m_textStream << "<PackingInformation\n";
			m_textStream << "fillingFactor  :  " << m_packingStructure.getFillingFactor() << "\n";
			m_textStream << "overlapVolume  :  " << " Exist \n";
			m_textStream << "PackingInformation>\n";
		}

		else
		{
			m_textStream << "<PackingInformation\n";
			m_textStream << "fillingFactor  :   " << m_packingStructure.getFillingFactor() << "\n";
			m_textStream << "overlapVolume  :   " << overlapVolume << "\n";
			m_textStream << "minimumZ       :  " << minimumZ << "\n";
			m_textStream << "PackingInformation>\n";
		}
	}
}
