#include "MakeXsfFile.h"

#include <map>

using namespace rk::dnspack;


MakeXsfFile::MakeXsfFile(const std::string& fileName, PackingStructure& packingStructure)
	: m_fileName{ fileName }
	, m_packingStructure(packingStructure)
	, m_basisComposition(packingStructure.getBasisComposition())
{
}

MakeXsfFile::~MakeXsfFile()
{
}

void MakeXsfFile::writePackingInfo()
{
	std::size_t sphereKindNumber = m_packingStructure.getSphereKindNumber();

	std::vector<std::string> sphereNames(6);
	{
		sphereNames[0] = "H";
		sphereNames[1] = "Na";
		sphereNames[2] = "Sc";
		sphereNames[3] = "Y";
		sphereNames[4] = "La";
		sphereNames[5] = "Hf";
	}

	std::vector<double> radii;
	{
		for (const auto& radius : m_basisComposition.getRadii())
		{
			radii.push_back(radius);
		}
	}

	if (6 < radii.size())
	{
		throw std::runtime_error("The kind of sphere is larger than 6.");
	}

	std::map<double, std::string> radiiAndName;
	{
		std::vector<std::string>::const_iterator nameIter = sphereNames.begin();

		for (const auto& radius : radii)
		{
			radiiAndName.emplace(radius, *nameIter);
			++nameIter;
		}
	}

	m_textStream << "#\n" << "# Spheres Infomation\n" << "#\n\n";

	m_textStream << "# <Definition.of.Sphere.Species\n";

	for (std::size_t label = 0; label < sphereKindNumber; ++label)
	{
		m_textStream << "# " << sphereNames[label] << "    " << radii[label] << "\n";
	}

	m_textStream << "# Definition.of.Sphere.Species>\n\n";


	m_textStream << "CRYSTAL" << "\n\n";

	const BravaisLattice& bravaisLattice = m_packingStructure.getBravaisLattice();

	m_textStream << "PRIMVEC\n";
	m_textStream << "  " << bravaisLattice[0].x << "  " << bravaisLattice[0].y << "  " << bravaisLattice[0].z << "\n";
	m_textStream << "  " << bravaisLattice[1].x << "  " << bravaisLattice[1].y << "  " << bravaisLattice[1].z << "\n";
	m_textStream << "  " << bravaisLattice[2].x << "  " << bravaisLattice[2].y << "  " << bravaisLattice[2].z << "\n";
	m_textStream << "PRIMCOORD\n";

	m_textStream << m_packingStructure.getSpheresNumber() << " 1\n";

	for (const auto& sphere : m_packingStructure.getBasis())
	{
		int iFrac;
		double Cart[3], Frac[3];

		Frac[0] = sphere.getFractionalCoordinate().a;
		Frac[1] = sphere.getFractionalCoordinate().b;
		Frac[2] = sphere.getFractionalCoordinate().c;
		for (iFrac = 0; iFrac < 3; iFrac++) Cart[iFrac] = 0.0;
		for (iFrac = 0; iFrac < 3; iFrac++) {
			Cart[0] += Frac[iFrac] * bravaisLattice[iFrac].x;
			Cart[1] += Frac[iFrac] * bravaisLattice[iFrac].y;
			Cart[2] += Frac[iFrac] * bravaisLattice[iFrac].z;
		}

		m_textStream << radiiAndName.at(sphere.getRadius()) << "  " << Cart[0] << "  " << Cart[1] << "  " << Cart[2] << "\n";
	}

	double overlapVolume = 0.0;
	double minimumZ = 0.0;
	{
		if (m_packingStructure.overlapExistence(overlapVolume, minimumZ))
		{
			m_textStream << "\n\n# <PackingInformation\n";
			m_textStream << "# fillingFactor  :" << m_packingStructure.getFillingFactor() << "\n";
			m_textStream << "# overlapVolume  :" << " Exist \n";
			m_textStream << "# PackingInformation>\n";
		}

		else
		{
			m_textStream << "\n\n# <PackingInformation\n";
			m_textStream << "# fillingFactor  :  " << m_packingStructure.getFillingFactor() << "\n";
			m_textStream << "# overlapVolume  :  " << overlapVolume << "\n";
			m_textStream << "# minimumZ       :  " << minimumZ << "\n";
			m_textStream << "# PackingInformation>\n";
		}
	}
}
