#include "PackingFactory.h"

#include <sstream>
#include <random>
#include <thread>

#include "RandomGenerate.h"
#include "InitialExpand.h"
#include "InitialArrange.h"
#include "SpherePackingOptimize.h"

#include "MakeDatFile.h"
#include "MakeXsfFile.h"
#include "ReportCurrentSituation.h"
#include "MakeResultFile.h"

using namespace rk::dnspack;

PackingFactory::PackingFactory(const BasisComposition& basisComposition, const PackingFactory::OptimizeParameters& optimizeParameters, const PackingFactory::DirectoryNames& directoryNames, PackingFactory::FactoryAttribute& factoryAttribute)
	: m_basisComposition(basisComposition)
	, m_optimizeParameters(optimizeParameters)
	, m_directoryNames(directoryNames)
	, m_factoryAttribute(factoryAttribute)
{
}

PackingFactory::~PackingFactory()
{
}


void PackingFactory::operator()(PackingFactory::MutexCluster& mutexCluster)
{
	const std::thread::id myThreadID{ std::this_thread::get_id() };

	CalculationAnalysis calculationAnalysis;

	ReportCurrentSituation reporter;
	{
		std::stringstream reportMessage;
		reportMessage << "Thread ID  =  " << myThreadID << "  :  Begin to produce packing structures. \n";
		reporter.report(reportMessage.str(), mutexCluster.reportMutex);
	}

	RejectNumbers rejectNumbers;
	MakeResultFile makeResultFile;


// ******************************************************************************************************************************************
	// packing produce

	while (m_factoryAttribute.createdNumber < m_factoryAttribute.creatPlanNumber)
	{
		std::random_device seed_gen;
		std::default_random_engine engine(seed_gen());

		std::uniform_real_distribution<double> initialArrangeDist(m_optimizeParameters.initialArrangeMaxChangeSizeLower, m_optimizeParameters.initialArrangeMaxChangeSizeUpper);
		std::uniform_real_distribution<double> spherePackingOptimizeDist(m_optimizeParameters.spherePackingOptimizeMaxChangeSizeLower, m_optimizeParameters.spherePackingOptimizeMaxChangeSizeUpper);

		RandomGenerate randomGenerate(m_basisComposition);
		PackingStructure packingStructure = randomGenerate();

		InitialExpand initialExpand(packingStructure, m_optimizeParameters.initialExpandMaxIteratorNumber, m_optimizeParameters.initialExpandMaxChangeSize);
		initialExpand();

		double initialArrangeMaxChangeSize = initialArrangeDist(engine);
		InitialArrange initialArrange(packingStructure, m_optimizeParameters.initialArrangeIteratorNumber, initialArrangeMaxChangeSize);

		if (!initialArrange())
		{
			std::lock_guard<std::mutex> lock{ mutexCluster.reportMutex };

			rejectNumbers.initialArrangeMinusVolumeRejectNumber += 1;
			m_factoryAttribute.createdNumber += 1;
			continue;
		}

		double spherePackingOptimizeMaxChangeSize = spherePackingOptimizeDist(engine);
		SpherePackingOptimize spherePackingOptimize(packingStructure, m_optimizeParameters.spherePackingOptimizeIteratorNumber, spherePackingOptimizeMaxChangeSize, m_optimizeParameters.spherePackingOptimizeDecreaseFactor);

		if (!spherePackingOptimize())
		{
			std::lock_guard<std::mutex> lock{ mutexCluster.reportMutex };

			rejectNumbers.spherePackingOptimizeMinusVolumeRejectNumber += 1;
			m_factoryAttribute.createdNumber += 1;
			continue;
		}

		else
		{
			double overlapVolume = 0.0;
			double minimumZ = 0.0;

			if (packingStructure.overlapExistence(overlapVolume, minimumZ))
			{
				std::lock_guard<std::mutex> lock{ mutexCluster.reportMutex };

				rejectNumbers.spherePackingOptimizeZvalueRejectNumber += 1;
				m_factoryAttribute.createdNumber += 1;
				continue;
			}

			else
			{
				if (1.0e-10 < overlapVolume)
				{
					std::lock_guard<std::mutex> lock{ mutexCluster.reportMutex };

					rejectNumbers.spherePackingOptimizeOverlapRejectNumber += 1;
					m_factoryAttribute.createdNumber += 1;
					continue;
				}

				else
				{
					double fillingFactor = packingStructure.getFillingFactor();

					std::lock_guard<std::mutex> lock{ mutexCluster.reportMutex };

					m_factoryAttribute.createdNumber += 1;

					{
						int label = static_cast<int>(100 * (fillingFactor - 0.60));

						if (0 <= label && label < 25)
						{
							calculationAnalysis.fillingFactorDistribution[label] += 1;
						}
					}

					if ((fillingFactor - m_factoryAttribute.maxFillingFactor) < 1.0e-6)
					{
						continue;
					}

					else
					{
						m_factoryAttribute.maxFillingFactor = fillingFactor;

						{
							int number = m_factoryAttribute.createdNumber + m_factoryAttribute.thresholdTrialNumber;

							if (m_factoryAttribute.maxTrialNumber < number)
							{
								number = m_factoryAttribute.maxTrialNumber;
							}

							m_factoryAttribute.creatPlanNumber = number;
						}

						if (fillingFactor < m_factoryAttribute.outputMinimumFillingFactor)
						{
							std::stringstream resultMessage;
							resultMessage << "Trial Label  =  " << m_factoryAttribute.createdNumber << "  :  Find a denser filling factor " << fillingFactor << "\n";
							makeResultFile.reportMessage(resultMessage.str());

							continue;
						}


						std::string datFileName = m_directoryNames.datFilesDirectoryName + std::to_string(fillingFactor) + ".dat";
						std::string xsfFileName = m_directoryNames.xsfFilesDirectoryName + std::to_string(fillingFactor) + ".xsf";

						MakeDatFile makeDatFile(datFileName, packingStructure);
						MakeXsfFile makeXsfFile(xsfFileName, packingStructure);

						std::stringstream versionMessage;
						versionMessage << "*******************************************************************************************\n\n";
						versionMessage << "The Version of SearchDensestPackings.exe  :  VERSION 0.0\n\n";
						versionMessage << "*******************************************************************************************\n";

						makeDatFile.appendMessage(versionMessage.str());

						makeDatFile.writePackingInfo();
						makeXsfFile.writePackingInfo();

						std::stringstream optimizeMessage;

						optimizeMessage << "\n\n#\n" << "# Creation Information\n" << "#\n\n\n";

						optimizeMessage << "Trial.Label                                " << m_factoryAttribute.createdNumber;


						optimizeMessage << "\n\n#\n" << "# Optimize Paramaters\n" << "#\n\n\n";
						optimizeMessage << "InitialExpand.MaxChangeSize                   " << m_optimizeParameters.initialExpandMaxChangeSize << "\n";
						optimizeMessage << "InitialExpand.Maximum.Iterator.Number         " << m_optimizeParameters.initialExpandMaxIteratorNumber << "\n";
						optimizeMessage << "InitialArrange.MaxChangeSize                  " << initialArrangeMaxChangeSize << "\n";
						optimizeMessage << "InitialArrange.Iterator.Number                " << m_optimizeParameters.initialArrangeIteratorNumber << "\n";
						optimizeMessage << "SpherePackingOptimize.MaxChangeSize           " << spherePackingOptimizeMaxChangeSize << "\n";
						optimizeMessage << "SpherePackingOptimize.Decrease.Factor         " << m_optimizeParameters.spherePackingOptimizeDecreaseFactor << "\n";
						optimizeMessage << "SpherePackingOptimize.Iterator.Number         " << m_optimizeParameters.spherePackingOptimizeIteratorNumber << "\n";

						makeDatFile.appendMessage(optimizeMessage.str());

						makeDatFile.makeFile();
						makeXsfFile.makeFile();


						{
							std::stringstream resultMessage;
							resultMessage << "Trial Label  =  " << m_factoryAttribute.createdNumber << "  :  Find a denser filling factor " << fillingFactor << "\n";
							makeResultFile.reportMessage(resultMessage.str());
						}
					}
				}
			}
		}
	}

	// packing produce
// ******************************************************************************************************************************************


	calculationAnalysis.appendRejectNumber(rejectNumbers, mutexCluster.reportMutex);
}
