#include "ReadInputFile.h"

#include <fstream>
#include <sstream>
#include <regex>
#include <map>
#include <thread>
#include <stdexcept>

using namespace rk::dnspack;

SphereInfo::SphereInfo()
	: compositionRatio{ 0 }
	, minimumRadius{ 0 }
	, maximumRadius{ 0 }
	, widthOfRadiusTicks{ 0 }
{
}

SphereInfo::SphereInfo(const int ratio, const double minRadius, const double maxRadius, const double width)
	: compositionRatio{ ratio }
	, minimumRadius{ minRadius }
	, maximumRadius{ maxRadius }
	, widthOfRadiusTicks{ width }
{
}

SphereInfo::~SphereInfo()
{
}


ReadInputFile::ReadInputFile(const std::string& fileName)
	: m_fileName{ fileName }
{
}

ReadInputFile::~ReadInputFile()
{
}

void ReadInputFile::initialize()
{
	std::string fileString;
	{
		std::ifstream inputFileStream;
		inputFileStream.open(m_fileName, std::ios_base::in);
		if (!inputFileStream)
		{
			throw std::runtime_error("could not open the input file.");
		}

		for (std::string stringLine; std::getline(inputFileStream, stringLine);)
		{
			fileString += stringLine + "\n";
		}
		inputFileStream.close();
	}


	{
		std::regex pat("[\\s]*Output.Directory.Name[\\s]+(.+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_outputDirectoryName = numSm.str(1);
		}

		else
		{
			throw std::runtime_error("couldn't get the output directory name.");
		}
	}

	std::map<std::string, int> nameAndComposition;
	{
		std::size_t initialPos = fileString.find("<Spheres.Composition.Ratio");
		std::size_t endPos = fileString.find("Spheres.Composition.Ratio>");

		if (initialPos != std::string::npos && endPos != std::string::npos)
		{
			std::size_t tempPos = fileString.find_first_of("\n", initialPos);
			std::string spheresString{ fileString, (tempPos + 1), (endPos - tempPos - 1) };
			std::stringstream spheresSStream{ spheresString };

			std::regex pat("[\\s]*([\\w]+)[\\s]+([\\d]+)[\\s]*");

			for (std::string line; std::getline(spheresSStream, line);)
			{
				std::smatch compositionSm;

				if (std::regex_match(line, compositionSm, pat))
				{
					nameAndComposition.emplace(compositionSm.str(1), std::stoi(compositionSm.str(2)));
				}

				else
				{
					throw std::runtime_error("couldn't get the spheres composition ratio.");
				}
			}
		}

		else
		{
			throw std::runtime_error("couldn't get the spheres composition ratio.");
		}
	}

	{
		std::size_t initialPos = fileString.find("<Radius.Range.And.Width.Of.Radius.Ticks");
		std::size_t endPos = fileString.find("Radius.Range.And.Width.Of.Radius.Ticks>");

		if (initialPos != std::string::npos && endPos != std::string::npos)
		{
			std::size_t tempPos = fileString.find_first_of("\n", initialPos);
			std::string spheresString{ fileString, (tempPos + 1), (endPos - tempPos - 1) };
			std::stringstream spheresSStream{ spheresString };

			std::regex pat("[\\s]*([\\w]+)[\\s]+([\\d.]+)[\\s]+([\\d.]+)[\\s]+([\\d.]+)[\\s]*");

			for (std::string line; std::getline(spheresSStream, line);)
			{
				std::smatch radiusSm;

				if (std::regex_match(line, radiusSm, pat))
				{
					SphereInfo sphereInfo;
					{
						sphereInfo.compositionRatio = nameAndComposition.at(radiusSm.str(1));
						sphereInfo.minimumRadius = std::stod(radiusSm.str(2));
						sphereInfo.maximumRadius = std::stod(radiusSm.str(3));
						sphereInfo.widthOfRadiusTicks = std::stod(radiusSm.str(4));
					}

					m_sphereInfo.push_back(std::move(sphereInfo));
				}

				else
				{
					throw std::runtime_error("couldn't get the radius info.");
				}
			}
		}

		else
		{
			throw std::runtime_error("couldn't get the radius info.");
		}
	}


	{
		std::regex pat("[\\s]*Minimum.Spheres.Number.Per.Unit.Cell[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_minimumSpheresNumber = std::stoi(numSm.str(1));
		}

		else
		{
			throw std::runtime_error("couldn't get the minimum spheres number per unit cell.");
		}
	}

	{
		std::regex pat("[\\s]*Maximum.Spheres.Number.Per.Unit.Cell[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_maximumSpheresNumber = std::stoi(numSm.str(1));
		}

		else
		{
			throw std::runtime_error("couldn't get the maximum spheres number per unit cell.");
		}
	}

	{
		std::regex pat("[\\s]*Threshold.Trial.Number.Of.Determine.Convergence[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_thresholdTrialNumber = std::stoi(numSm.str(1));
		}

		else
		{
			throw std::runtime_error("couldn't get the trial number.");
		}
	}

	{
		std::regex pat("[\\s]*Maximum.Trial.Number.Of.Each.Composition[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_maxTrialNumber = std::stoi(numSm.str(1));
		}

		else
		{
			throw std::runtime_error("couldn't get the trial number.");
		}
	}

	{
		std::regex pat("[\\s]*InitialExpand.MaxChangeSize[\\s]+([\\d.]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_initialExpandMaxChangeSize = std::stod(numSm.str(1));
		}

		else
		{
			m_initialExpandMaxChangeSize = 0.3;
		}
	}

	{
		std::regex pat("[\\s]*InitialExpand.Maximum.Iterator.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_initialExpandMaxIteratorNumber = std::stoi(numSm.str(1));
		}

		else
		{
			m_initialExpandMaxIteratorNumber = 500;
		}
	}

	{
		std::regex pat("[\\s]*InitialArrange.MaxChangeSize.Upper[\\s]+([\\d.]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_initialArrangeMaxChangeSizeUpper = std::stod(numSm.str(1));
		}

		else
		{
			m_initialArrangeMaxChangeSizeUpper = 0.15;
		}
	}

	{
		std::regex pat("[\\s]*InitialArrange.MaxChangeSize.Lower[\\s]+([\\d.]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_initialArrangeMaxChangeSizeLower = std::stod(numSm.str(1));
		}

		else
		{
			m_initialArrangeMaxChangeSizeLower = 0.03;
		}
	}

	{
		std::regex pat("[\\s]*InitialArrange.Iterator.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_initialArrangeIteratorNumber = std::stoi(numSm.str(1));
		}

		else
		{
			m_initialArrangeIteratorNumber = 2000;
		}
	}

	{
		std::regex pat("[\\s]*SpherePackingOptimize.MaxChangeSize.Upper[\\s]+([\\d.]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_spherePackingOptimizeMaxChangeSizeUpper = std::stod(numSm.str(1));
		}

		else
		{
			m_spherePackingOptimizeMaxChangeSizeUpper = 0.05;
		}
	}

	{
		std::regex pat("[\\s]*SpherePackingOptimize.MaxChangeSize.Lower[\\s]+([\\d.]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_spherePackingOptimizeMaxChangeSizeLower = std::stod(numSm.str(1));
		}

		else
		{
			m_spherePackingOptimizeMaxChangeSizeLower = 0.02;
		}
	}

	{
		std::regex pat("[\\s]*SpherePackingOptimize.Decrease.Factor[\\s]+([\\d.]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_spherePackingOptimizeDecreaseFactor = std::stod(numSm.str(1));
		}

		else
		{
			m_spherePackingOptimizeDecreaseFactor = 0.9997;
		}
	}

	{
		std::regex pat("[\\s]*SpherePackingOptimize.Iterator.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_spherepackingOptimizeIteratorNumber = std::stoi(numSm.str(1));
		}

		else
		{
			m_spherepackingOptimizeIteratorNumber = 40000;
		}
	}

	{
		std::regex pat("[\\s]*Thread.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_threadNumber = std::stoi(numSm.str(1));
		}

		else
		{
			m_threadNumber = static_cast<int>(std::thread::hardware_concurrency());
		}
	}

	{
		std::regex pat("[\\s]*Output.Minimum.FillingFactor[\\s]+([\\d.]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_outputMinimumFillingFactor = std::stod(numSm.str(1));
		}

		else
		{
			m_initialArrangeMaxChangeSizeUpper = 0.7;
		}
	}


	{
		bool maximumSphereExist = false;

		for (const auto& info : m_sphereInfo)
		{
			if (info.minimumRadius == 1.0 && info.maximumRadius == 1.0 && info.widthOfRadiusTicks == 0.0)
			{
				maximumSphereExist = true;
				continue;
			}
		}

		if (!maximumSphereExist)
		{
			throw std::runtime_error("Maximum Sphere does not exist.");
		}
	}

	for (const auto& info : m_sphereInfo)
	{
		if (info.compositionRatio < 1)
		{
			throw std::runtime_error("The composition ratio is less than 1.");
		}

		if (info.minimumRadius == 1.0 && info.maximumRadius == 1.0 && info.widthOfRadiusTicks == 0.0)
		{
			continue;
		}

		else
		{
			if (info.maximumRadius < info.minimumRadius)
			{
				throw std::runtime_error("The maximum radius is less minimum radius.");
			}

			if (1.0 <= info.maximumRadius)
			{
				throw std::runtime_error("The maximum radius is more than 1.0.");
			}

			if (info.minimumRadius <= 0.0)
			{
				throw std::runtime_error("The minimum radius is less than 0.0.");
			}

			if (info.widthOfRadiusTicks < 0.0001)
			{
				throw std::runtime_error("Width of radius ticks is less than 0.0001.");
			}
		}
	}


	if (nameAndComposition.size() != m_sphereInfo.size())
	{
		throw std::runtime_error("Couldn't get basis information correctly.");
	}

	if (m_minimumSpheresNumber < 1)
	{
		throw std::runtime_error("Minimum.Spheres.Number.Per.Unit.Cell is less than 1.");
	}

	if (m_maximumSpheresNumber < m_minimumSpheresNumber)
	{
		throw std::runtime_error("Minimum.Spheres.Number.Per.Unit.Cell is larger than Maximum.Spheres.Number.Per.Unit.Cell.");
	}

	if (m_maxTrialNumber < m_thresholdTrialNumber)
	{
		throw std::runtime_error("Threshold.Trial.Number.Of.Determine.Convergence is more than Maximum.Trial.Number.Of.Each.Composition");
	}

	if (m_thresholdTrialNumber < 1)
	{
		throw std::runtime_error("Threshold.Trial.Number.Of.Determine.Convergence is less than one.");
	}

	if (m_maxTrialNumber < 1)
	{
		throw std::runtime_error("Maximum.Trial.Number.Of.Each.Composition is less than one.");
	}

	if (m_initialArrangeMaxChangeSizeUpper < m_initialArrangeMaxChangeSizeLower)
	{
		throw std::runtime_error("InitialArrange.MaxChangeSize.Upper is less than InitialArrange.MaxChangeSize.Lower.");
	}

	if (m_spherePackingOptimizeMaxChangeSizeUpper < m_spherePackingOptimizeMaxChangeSizeLower)
	{
		throw std::runtime_error("SpherePackingOptimize.MaxChangeSize.Upper is less than SpherePackingOptimize.MaxChangeSize.Lower.");
	}

	if (0.2 < m_initialArrangeMaxChangeSizeUpper)
	{
		throw std::runtime_error("InitialArrange.MaxChangeSize.Upper is more than 0.2.");
	}

	if (0.2 < m_spherePackingOptimizeMaxChangeSizeUpper)
	{
		throw std::runtime_error("SpherePackingOptimize.MaxChangeSize.Upper is more than 0.2.");
	}

	if (m_threadNumber < 1)
	{
		throw std::runtime_error("Thread.Number is less than 1.");
	}

	if (m_outputMinimumFillingFactor < 0.0)
	{
		throw std::runtime_error("Output.Minimum.FillingFactor is less than 0.0.");
	}
}
