#include "SetTrialBasisComposition.h"

#include <set>
#include <map>
#include <iterator>
#include <algorithm>

using namespace rk::dnspack;

SetTrialBasisComposition::SetTrialBasisComposition(const std::vector<SphereInfo>& sphereInfo, const int minimumSpheresNumber, int maximumSpheresNumber)
	: m_sphereInfo{ sphereInfo }
	, m_minimumSpheresNumber{ minimumSpheresNumber }
	, m_maximumSpheresNumber{ maximumSpheresNumber }
{
}

SetTrialBasisComposition::~SetTrialBasisComposition()
{
}

void SetTrialBasisComposition::operator()(std::vector<BasisComposition>& trialBasisComposition)
{
	trialBasisComposition.clear();

	int sphereKindNumber = static_cast<int>(m_sphereInfo.size());

	std::vector<std::vector<int>> trialCompositionRatio;

	int baseSpheresNumber = 0;
	std::vector<int> baseCompositionRatio;
	{
		for (const auto& info : m_sphereInfo)
		{
			baseCompositionRatio.push_back(info.compositionRatio);
			baseSpheresNumber += info.compositionRatio;
		}
	}

	int sphersNumber = baseSpheresNumber;
	int scale = 1;
	{
		for (; sphersNumber < m_minimumSpheresNumber; ++scale)
		{
			sphersNumber += baseSpheresNumber;
		}
	}

	for (; sphersNumber <= m_maximumSpheresNumber; ++scale)
	{
		std::vector<int> compositionRatio;

		for (const auto& info : m_sphereInfo)
		{
			compositionRatio.push_back(info.compositionRatio  * scale);
		}

		trialCompositionRatio.push_back(std::move(compositionRatio));

		sphersNumber += baseSpheresNumber;
	}


	std::vector<std::vector<double>> radiiCandidates;
	{
		for (const auto& info : m_sphereInfo)
		{
			std::vector<double> candidates;

			if (1.0 == info.minimumRadius && 1.0 == info.maximumRadius && 0.0 == info.widthOfRadiusTicks)
			{
				candidates.push_back(1.0);
				radiiCandidates.push_back(std::move(candidates));
			}

			else
			{
				double radius = info.minimumRadius;

				while ((radius - info.maximumRadius) < 0.0001)
				{
					candidates.push_back(radius);

					radius += info.widthOfRadiusTicks;
				}

				radiiCandidates.push_back(std::move(candidates));
			}
		}
	}

	std::vector<std::vector<double>> trialRadii;
	{
		std::vector<std::vector<double>::const_iterator> candidateIters(sphereKindNumber);
		std::vector<std::vector<double>::const_iterator> candidateIterEnds(sphereKindNumber);

		for (int label = 0; label < sphereKindNumber; ++label)
		{
			candidateIters[label] = radiiCandidates[static_cast<std::size_t>(label)].cbegin();
			candidateIterEnds[label] = radiiCandidates[static_cast<std::size_t>(label)].cend();
		}

		{
			std::vector<double> radii;

			for (const auto& iter : candidateIters)
			{
				radii.push_back(*iter);
			}

			if (std::is_sorted(radii.begin(), radii.end()))
			{
				std::vector<double>::const_iterator iter = radii.cbegin();
				std::vector<double>::const_iterator iterPrevEnd = std::prev(radii.cend());

				bool push = true;

				while (iter != iterPrevEnd)
				{
					double radius1 = *iter;
					++iter;
					double radius2 = *iter;

					double difference = radius2 - radius1;

					if (difference < 0.0001)
					{
						push = false;
						break;
					}
				}

				if (push)
				{
					trialRadii.push_back(std::move(radii));
				}
			}
		}

		for (bool setSuccess = true; setSuccess == true; )
		{
			setSuccess = false;

			for (int label = 0; label < sphereKindNumber; ++label)
			{
				++candidateIters[label];

				if (candidateIters[label] == candidateIterEnds[label])
				{
					candidateIters[label] = radiiCandidates[static_cast<std::size_t>(label)].cbegin();
					continue;
				}

				else
				{
					setSuccess = true;
					break;
				}
			}

			if (setSuccess)
			{
				std::vector<double> radii;

				for (const auto& iter : candidateIters)
				{
					radii.push_back(*iter);
				}

				if (std::is_sorted(radii.begin(), radii.end()))
				{
					std::vector<double>::const_iterator iter = radii.cbegin();
					std::vector<double>::const_iterator iterPrevEnd = std::prev(radii.cend());

					bool push = true;

					while (iter != iterPrevEnd)
					{
						double radius1 = *iter;
						++iter;
						double radius2 = *iter;

						double difference = radius2 - radius1;

						if (difference < 0.0001)
						{
							push = false;
							break;
						}
					}

					if (push)
					{
						trialRadii.push_back(std::move(radii));
					}
				}
			}
		}
	}


	for (const auto& compositionRatio : trialCompositionRatio)
	{
		for (const auto& radiiSet : trialRadii)
		{
			std::set<double> radii;
			std::map<double, int> numbers;

			for (int label = 0; label < sphereKindNumber; ++label)
			{
				radii.insert(radiiSet.at(static_cast<std::size_t>(label)));
				numbers.emplace(radiiSet.at(label), compositionRatio.at(static_cast<std::size_t>(label)));
			}

			trialBasisComposition.push_back(BasisComposition(std::move(radii), std::move(numbers)));
		}
	}
}
