#include "SpherePackingOptimize.h"

#include <algorithm>
#include <numeric>

using namespace rk::dnspack;

SpherePackingOptimize::SpherePackingOptimize(PackingStructure& packingStructure, const int maxStepNumber, const double maxChangeSize, const double decreaseFactor)
	: m_packingStructure(packingStructure)
	, m_spheresNumber{ static_cast<int>(packingStructure.getSpheresNumber()) }
	, m_maxStepNumber{ maxStepNumber }
	, m_maxChangeSize{ maxChangeSize }
	, m_decreaseFactor{ decreaseFactor }
{
}

SpherePackingOptimize::~SpherePackingOptimize()
{
}

bool SpherePackingOptimize::setNeighborSpheres()
{
	m_neighborSpheres.clear();
	m_neighborSpheres.reserve(12 * static_cast<size_t>(m_spheresNumber));

	const std::vector<Sphere>& basis(m_packingStructure.getBasis());
	const BravaisLattice& bravaisLattice(m_packingStructure.getBravaisLattice());

	for (auto& sphere : m_packingStructure.getBasis())
	{
		sphere.refineFractionalCoordinate();
	}

	int Q[3];
	{
		CartesianCoordinate reciprocalLatticeVector[3];
		{
			double volume = bravaisLattice.getVolume();

			reciprocalLatticeVector[0] = crossProduct(bravaisLattice[1], bravaisLattice[2]);
			reciprocalLatticeVector[0].scaling((1 / volume));

			reciprocalLatticeVector[1] = crossProduct(bravaisLattice[2], bravaisLattice[0]);
			reciprocalLatticeVector[1].scaling((1 / volume));

			reciprocalLatticeVector[2] = crossProduct(bravaisLattice[0], bravaisLattice[1]);
			reciprocalLatticeVector[2].scaling((1 / volume));
		}

		Q[0] = static_cast<int>(2.1 * reciprocalLatticeVector[0].getNorm()) + 1;
		Q[1] = static_cast<int>(2.1 * reciprocalLatticeVector[1].getNorm()) + 1;
		Q[2] = static_cast<int>(2.1 * reciprocalLatticeVector[2].getNorm()) + 1;
	}

	{
		if (3 < Q[0])
		{
			return false;
		}

		if (3 < Q[1])
		{
			return false;
		}

		if (3 < Q[2])
		{
			return false;
		}
	}

	for (int baseIndex = 0; baseIndex < m_spheresNumber; ++baseIndex)
	{
		for (int Qa = -Q[0]; Qa <= Q[0]; ++Qa)
		{
			for (int Qb = -Q[1]; Qb <= Q[1]; ++Qb)
			{
				for (int Qc = -Q[2]; Qc <= Q[2]; ++Qc)
				{
					FractionalCoordinate Q{ static_cast<double>(Qa), static_cast<double>(Qb), static_cast<double>(Qc) };

					for (int sphereIndex = (baseIndex + 1); sphereIndex < m_spheresNumber; ++sphereIndex)
					{
						FractionalCoordinate relativeFractionalVector;
						relativeFractionalVector = basis[sphereIndex].getFractionalCoordinate() + Q - basis[baseIndex].getFractionalCoordinate();

						CartesianCoordinate relativeVector;
						{
							relativeVector += scaling(relativeFractionalVector.a, bravaisLattice[0]);
							relativeVector += scaling(relativeFractionalVector.b, bravaisLattice[1]);
							relativeVector += scaling(relativeFractionalVector.c, bravaisLattice[2]);
						}

						double spheresDistanceSquare = relativeVector.getNormSquare();

						double radiusSum = basis[sphereIndex].getRadius() + basis[baseIndex].getRadius();
						double threshold = 0.5 * (0.5 + 2 * radiusSum);

						if ((spheresDistanceSquare - radiusSum * radiusSum) < threshold)
						{
							SphereLabel sphereLabel{ Qa,Qb,Qc,sphereIndex };
							m_neighborSpheres.push_back(std::pair<int, SphereLabel>(baseIndex, sphereLabel));
						}
					}

					if (!(Qa == 0 && Qb == 0 && Qc == 0))
					{
						CartesianCoordinate relativeVector(0, 0, 0);
						{
							relativeVector += scaling(Q.a, bravaisLattice[0]);
							relativeVector += scaling(Q.b, bravaisLattice[1]);
							relativeVector += scaling(Q.c, bravaisLattice[2]);
						}

						double spheresDistanceSquare = relativeVector.getNormSquare();

						double radiusSum = 2 * basis[baseIndex].getRadius();
						double threshold = 0.5 * (0.5 + 2 * radiusSum);

						if ((spheresDistanceSquare - radiusSum * radiusSum) < threshold)
						{
							SphereLabel sphereLabel{ Qa,Qb,Qc,baseIndex };
							m_neighborSpheres.push_back(std::pair<int, SphereLabel>(baseIndex, sphereLabel));
						}
					}
				}
			}
		}
	}

	m_neighborSpheres.shrink_to_fit();

	return true;
}

void SpherePackingOptimize::calculateGradient()
{
	m_gradientVector.resize(3 * static_cast<size_t>(m_spheresNumber) + 9, 0.0);

	const std::vector<Sphere>& basis(m_packingStructure.getBasis());
	const BravaisLattice& bravaisLattice(m_packingStructure.getBravaisLattice());

	for (const auto& neighborSpherePair : m_neighborSpheres)
	{
		int baseIndex = neighborSpherePair.first;
		int sphereIndex = neighborSpherePair.second.sphereIndex;

		FractionalCoordinate Q;
		{
			Q.a = static_cast<double>(neighborSpherePair.second.Qa);
			Q.b = static_cast<double>(neighborSpherePair.second.Qb);
			Q.c = static_cast<double>(neighborSpherePair.second.Qc);
		}

		FractionalCoordinate relativeFractionalVector;
		relativeFractionalVector = basis[sphereIndex].getFractionalCoordinate() + Q - basis[baseIndex].getFractionalCoordinate();

		CartesianCoordinate relativeVector;
		{
			relativeVector += scaling(relativeFractionalVector.a, bravaisLattice[0]);
			relativeVector += scaling(relativeFractionalVector.b, bravaisLattice[1]);
			relativeVector += scaling(relativeFractionalVector.c, bravaisLattice[2]);
		}

		double distanceSquare = relativeVector.getNormSquare();
		double radiusSum = basis[sphereIndex].getRadius() + basis[baseIndex].getRadius();

		if ((distanceSquare - radiusSum * radiusSum) < 0)
		{
			double distance = std::sqrt(distanceSquare);
			double zValue = distance - radiusSum;

			if (baseIndex != sphereIndex)
			{
				for (int direction = 0; direction < 3; ++direction)
				{
					int index = 3 * baseIndex + direction;

					double differential = ((-1) * innerProduct(bravaisLattice[direction], relativeVector)) / distance;
					m_gradientVector[index] += differential * force(zValue);
				}

				for (int direction = 0; direction < 3; ++direction)
				{
					int index = 3 * sphereIndex + direction;

					double differential = (innerProduct(bravaisLattice[direction], relativeVector)) / distance;
					m_gradientVector[index] += differential * force(zValue);
				}

				for (int latticeIndex = 0; latticeIndex < 3; ++latticeIndex) // relativeFractionalVector2|̂ŕ̕ω͕svB
				{
					for (int latticeDirection = 0; latticeDirection < 3; ++latticeDirection)
					{
						int index = 3 * m_spheresNumber + 3 * latticeIndex + latticeDirection;

						double differential = 0;
						{
							for (int s = 0; s < 3; ++s)
							{
								differential += bravaisLattice[s][latticeDirection] * relativeFractionalVector[s] * relativeFractionalVector[latticeIndex];
							}
							differential /= distance;
						}

						m_gradientVector[index] += (differential * force(zValue)); // 2
					}
				}
			}

			else
			{
				for (int latticeIndex = 0; latticeIndex < 3; ++latticeIndex)
				{
					for (int latticeDirection = 0; latticeDirection < 3; ++latticeDirection)
					{
						int index = 3 * m_spheresNumber + 3 * latticeIndex + latticeDirection;

						double differential = 0;
						{
							for (int s = 0; s < 3; ++s)
							{
								differential += bravaisLattice[s][latticeDirection] * relativeFractionalVector[s] * relativeFractionalVector[latticeIndex];
							}

							differential /= distance;
						}

						m_gradientVector[index] += (differential * force(zValue)) / 2;
					}
				}
			}
		}
	}

	double differentialOmega[3][3];
	{
		differentialOmega[0][0] = bravaisLattice[1].y * bravaisLattice[2].z - bravaisLattice[1].z * bravaisLattice[2].y;
		differentialOmega[0][1] = bravaisLattice[1].z * bravaisLattice[2].x - bravaisLattice[1].x * bravaisLattice[2].z;
		differentialOmega[0][2] = bravaisLattice[1].x * bravaisLattice[2].y - bravaisLattice[1].y * bravaisLattice[2].x;
		differentialOmega[1][0] = bravaisLattice[2].y * bravaisLattice[0].z - bravaisLattice[2].z * bravaisLattice[0].y;
		differentialOmega[1][1] = bravaisLattice[2].z * bravaisLattice[0].x - bravaisLattice[2].x * bravaisLattice[0].z;
		differentialOmega[1][2] = bravaisLattice[2].x * bravaisLattice[0].y - bravaisLattice[2].y * bravaisLattice[0].x;
		differentialOmega[2][0] = bravaisLattice[0].y * bravaisLattice[1].z - bravaisLattice[0].z * bravaisLattice[1].y;
		differentialOmega[2][1] = bravaisLattice[0].z * bravaisLattice[1].x - bravaisLattice[0].x * bravaisLattice[1].z;
		differentialOmega[2][2] = bravaisLattice[0].x * bravaisLattice[1].y - bravaisLattice[0].y * bravaisLattice[1].x;
	}

	for (int latticeKind = 0; latticeKind < 3; ++latticeKind)
	{
		for (int direction = 0; direction < 3; ++direction)
		{
			int index = 3 * m_spheresNumber + 3 * latticeKind + direction;

			m_gradientVector[index] += m_pressure * differentialOmega[latticeKind][direction];
		}
	}
}

void SpherePackingOptimize::optimize()
{
	double maxGradientSize = 0.0;
	{
		double maxGradientSizeSquare = 0.0;
		const BravaisLattice& bravaisLattice(m_packingStructure.getBravaisLattice());

		for (std::size_t sphereLabel = 0; sphereLabel < m_spheresNumber; ++sphereLabel)
		{
			std::size_t label = 3 * sphereLabel;

			CartesianCoordinate changeVector;
			{
				changeVector += scaling(m_gradientVector[label], bravaisLattice[0]);
				changeVector += scaling(m_gradientVector[label + 1], bravaisLattice[1]);
				changeVector += scaling(m_gradientVector[label + 2], bravaisLattice[2]);
			}

			double changeSquare = changeVector.x * changeVector.x + changeVector.y * changeVector.y + changeVector.z * changeVector.z;

			if (maxGradientSizeSquare < changeSquare)
			{
				maxGradientSizeSquare = changeSquare;
			}
		}

		for (std::size_t latticeLabel = 0; latticeLabel < 3; ++latticeLabel)
		{
			std::size_t label = 3 * static_cast<std::size_t>(m_spheresNumber) + latticeLabel;

			double changeX = m_gradientVector[label];
			double changeY = m_gradientVector[label + 1];
			double changeZ = m_gradientVector[label + 2];

			double changeSquare = changeX * changeX + changeY * changeY + changeZ * changeZ;

			if (maxGradientSizeSquare < changeSquare)
			{
				maxGradientSizeSquare = changeSquare;
			}
		}

		maxGradientSize = sqrt(maxGradientSizeSquare);
	}

	m_gradientVector *= (m_maxChangeSize / maxGradientSize);

	int label = 0;

	for (auto& sphere : m_packingStructure.getBasis())
	{
		sphere.getFractionalCoordinate().a -= m_gradientVector[label];
		++label;
		sphere.getFractionalCoordinate().b -= m_gradientVector[label];
		++label;
		sphere.getFractionalCoordinate().c -= m_gradientVector[label];
		++label;
	}

	BravaisLattice& bravaisLattice(m_packingStructure.getBravaisLattice());
	{
		for (int i = 0; i < 3; ++i)
		{
			bravaisLattice[i].x -= m_gradientVector[label];
			++label;
			bravaisLattice[i].y -= m_gradientVector[label];
			++label;
			bravaisLattice[i].z -= m_gradientVector[label];
			++label;
		}
	}
}
